import React, { useMemo, useRef, useState } from 'react'
import _ from 'lodash'
import Popover from '../Popover'
import {FilterPanel} from '../Grid/FilterPanel'
import AddFilterIcon from './../../assets/svgs/add-filter-icon.svg'
import CloseIcon from './../../assets/svgs/close-icon.svg'
import CaretIcon from './../../assets/svgs/triangle.svg'
import './ExternalTableFilters.scss'
import ButtonBase from '../ButtonBase'
import useEvent from '../../hooks/useEvent'

const ExternalColumnFilterButton = ({ columnId, label, columnFilterOptions, columnFilterState = { activeCount: 0 }, setColumnFilterInvisible, updateColumnFilter }) => {
  const [isOpen, setIsOpen] = useState(columnFilterState.activeCount === 0)
  const count = columnFilterState.activeCount

  const caretStyle = isOpen ? { transform: 'scaleY(-1)' } : undefined

  const ref = useRef()

  useEvent('click', (e) => {
    if (!ref?.current?.contains(e.target)) {
      setIsOpen(false)
    }
  }, true)

  const column = {
    filterOptions: columnFilterOptions,
    filterState: columnFilterState,
    updateColumnFilter: (e) => { updateColumnFilter(e); setIsOpen(false) }
  }

  return <div className="Suc-external-column-filter-button" ref={ref}>
    <div className={`Suc-extrenal-filter-first-line ${isOpen ? 'open' : ''}`}>
      <div className="Suc-filter-label" onClick={() => setIsOpen(!isOpen)}>{label}{count > 0 && ` (${count})`}</div>
      <div className="Suc-filter-actions">
      <ButtonBase onClick={() => setIsOpen(!isOpen)} className="Suc-filter-caret"><CaretIcon style={caretStyle}/></ButtonBase>
      <div className="Suc-filter-pipe">|</div>
      <ButtonBase className="Suc-filter-close" onClick={() => {
        setColumnFilterInvisible(columnId)
        updateColumnFilter({})
      }}>
        <CloseIcon/>
      </ButtonBase>
      </div>
    </div>
    {isOpen && <FilterPanel column={column} setIsOpen={setIsOpen}/>}
  </div>
}

const AddFilterPopover = ({ unfilteredColumns, setColumnFilterVisible }) => {
  const [isOpen, setIsOpen] = useState(false)
  const close = () => setIsOpen(false)

  return <Popover
    trigger={'click'}
    appendToBody={true}
    container={<AddFilterIcon/>}
    visible={isOpen}
    onVisibleChange={setIsOpen}
  >
    <div className="Suc-select-columns-popover">
      {unfilteredColumns.map(({ id, Header }) => {
        return <ButtonBase key={id} onClick={() => { setColumnFilterVisible(id); close() }}>
          {Header}
        </ButtonBase>
      })}
    </div>
  </Popover>
}

const ExternalTableFilters = ({ filterState, updateFilterState, columns, resetFilters }) => {
  const [visibleFilters, setVisibleFilters] = useState(Object.keys(filterState.columns))
  const columnsMap = _.keyBy(columns, 'id')
  const filterOptions = useMemo(() => columns.reduce((acc, col) => {
    if (col.filterOptions) {
      return { ...acc, [col.id]: col.filterOptions }
    } else return acc
  }, {}), [columns])

  const setColumnFilterVisible = (id) => {
    setVisibleFilters([...visibleFilters, id])
  }

  const setColumnFilterInvisible = (id) => {
    setVisibleFilters(visibleFilters.filter(fId => fId !== id))
  }

  const clearAll = () => {
    resetFilters()
    setVisibleFilters([])
  }

  const unfilteredColumns = useMemo(() => {
    return _.filter(columns, col => filterOptions[col.id] && !visibleFilters.find(f => f === col.id))
  }, [filterOptions, visibleFilters])

  return <div className="Suc-external-column-filters">
        <div className="Suc-external-filters-label">Filter By</div>
        <div className="Suc-external-filters-group">
          {visibleFilters.map(columnId => {
            const updateColumnFilter = (values) => {
              updateFilterState({
                id: columnId,
                values
              })
            }

            return <ExternalColumnFilterButton
              key={columnId}
              label={columnsMap[columnId].Header}
              columnFilterState={filterState.columns[columnId]}
              columnFilterOptions={filterOptions[columnId]}
              filterConfig={columnsMap[columnId].filter}
              columnId={columnId}
              setColumnFilterInvisible={setColumnFilterInvisible}
              updateColumnFilter={updateColumnFilter}
            />
          })}
        </div>
        <div className={`Suc-add-extrenal-filter ${unfilteredColumns.length === 0 ? 'disabled' : ''}`}><AddFilterPopover
            setColumnFilterVisible={setColumnFilterVisible}
            unfilteredColumns={unfilteredColumns}
        /></div>
        {filterState?.count > 0 && <ButtonBase className="Suc-external-filters-clear-all" onClick={clearAll}>Clear All</ButtonBase>}
    </div>
}

export default ExternalTableFilters
