import React, { useCallback, useMemo, useReducer } from 'react'
import _ from 'lodash'
import { SELECTION_TYPE, SelectionCell } from './Selection'
import useRunOnceWhenTruthy from '../../hooks/useRunOnceWhenTruthy'

const noFilterState = { columns: {}, count: 0, textFilter: '' }

const sort = (sortState, columnsMap, rows) => {
  return _.orderBy(rows, [columnsMap[sortState.column].accessor], [sortState.sortOrder])
}

const filter = (filterState, rows, columns, columnsMap) => {
  const { textFilter } = filterState
  return _.filter(rows, (row) => {
    const lowerCaseTextFilter = textFilter && textFilter.toLowerCase()
    let isTextFilterMatch = !textFilter

    if (textFilter) {
      _.forEach(columns, (column) => {
        const accessor = column.accessor;
        const value = _.get(row, accessor);

        if (_.isString(value) && value.toLowerCase().includes(lowerCaseTextFilter)) {
          isTextFilterMatch = true
          return false
        } else if (_.isArray(value) && value.join(', ').toLowerCase().includes(lowerCaseTextFilter)) {
          isTextFilterMatch = true
          return false
        }
      })
    }

    let isMultiFilterMatch = true

    _.forEach(filterState.columns, (filter, key) => {
      if (_.isObject(filter) && filter.activeCount > 0) {
        const accessor = columnsMap[key].accessor;
        const rowAccessor = _.get(row, accessor)
        if (_.isArray(rowAccessor)) {
          const existsOne = _.find(rowAccessor, value => {
            return filter.values[value]
          })
          if (!existsOne) {
            isMultiFilterMatch = false
            return false
          }
        } else {
          if (!filter.values[rowAccessor]) {
            isMultiFilterMatch = false
            return false
          }
        }
      }
    })

    return isTextFilterMatch && isMultiFilterMatch
  })
}

const init = (initialState) => {
  let sortState = null

  if (initialState.isSorting) {
    sortState = {
      column: _.isString(initialState.columns[0]) ? initialState.columns[0] : (initialState.columns[0].id || '0'),
      sortOrder: 'asc'
    }
  }

  const initialPageState = initialState.pageSize ? { pageIndex: 0 } : undefined

  return {
    sortState: initialState.initialSortState || sortState,
    initialSortState: initialState.initialSortState || sortState,
    filterState: initialState.initialFilterState || noFilterState,
    initialFilterState: initialState.initialFilterState || noFilterState,
    tableState: { ...initialState.defaultTableState } || {},
    rowState: {},
    columnState: {},
    selectionState: initialState.selectionType !== SELECTION_TYPE.NONE ? { rows: {}, count: 0 } : undefined,
    selectionType: initialState.selectionType,
    pageState: initialPageState,
    initialPageState
  }
}

const reducer = (state, action) => {
  const { type, payload } = action

  switch (type) {
    case 'TOGGLE-SORT': {
      const { id } = payload

      let newSortState = {}
      const currentSortOrder = state.sortState.sortOrder
      const currentSortColumn = state.sortState.column
      if (currentSortColumn === id) {
        if (currentSortOrder === 'asc') {
          newSortState.sortOrder = 'desc'
          newSortState.column = id
        } else if (currentSortColumn !== state.initialSortState.column) {
          newSortState = state.initialSortState
        } else {
          newSortState.sortOrder = 'asc'
          newSortState.column = id
        }
      } else {
        newSortState.sortOrder = 'asc'
        newSortState.column = id
      }

      return { ...state, sortState: newSortState, pageState: state.initialPageState }
    }
    case 'SET-FILTER': {
      const { id, values } = payload

      const newValues = _.pickBy(values)

      const newColumnFilter = {
        values: newValues,
        valuesArray: _.values(values),
        activeCount: _.size(newValues)
      }

      const columns = { ...state.filterState.columns, [id]: newColumnFilter }
      let count = 0

      _.forEach(columns, (state) => {
        if (state.activeCount > 0) {
          count = count + 1
        }
      })

      return { ...state, filterState: { ...state.filterState, columns, count }, pageState: state.initialPageState }
    }
    case 'RESET-FILTERS': {
      return {
        ...state,
        filterState: noFilterState,
        pageState: state.initialPageState
      }
    }
    case 'TOGGLE-ROW-SELECTION': {
      const { id, value } = payload
      let count = state.selectionState.count
      let updatedValue = value
      let oldSelectedRows = state.selectionState.rows
      if (state.selectionType === SELECTION_TYPE.MULTIPLE) {
        const currentValue = state.selectionState.rows[id] || false
        if (currentValue && !value) {
          count = count - 1
        } else if (!currentValue && value) {
          count = count + 1
        }
      }
      if (state.selectionType === SELECTION_TYPE.SINGULAR) {
        count = 1
        updatedValue = true
        oldSelectedRows = {}
      }

      return { ...state, selectionState: { count, rows: { ...oldSelectedRows, [id]: updatedValue } } }
    }
    case 'TOGGLE-SELECT-ALL': {
      const { value, rows } = payload
      const selectionState = { rows: {}, count: 0, allSelected: value }
      if (value) {
        _.forEach(rows, (row) => {
          selectionState.rows[row.id] = value
        })
        selectionState.count = rows.length
      }
      return { ...state, selectionState }
    }
    case 'UPDATE-TEXT-FILTER': {
      const { value } = payload
      return { ...state, filterState: { ...state.filterState, textFilter: value }, pageState: state.initialPageState }
    }
    case 'GO-TO-PAGE': {
      return { ...state, pageState: { pageIndex: payload.index } }
    }
  }

  return state
}

export default ({ isSorting, selectionType = SELECTION_TYPE.NONE, columns, rows: data, externalSort, externalFilter, initialSortState, initialFilterState, pageSize = 50, defaultSelectedRows }) => {
  const { pColumns, columnsMap } = useMemo(() => {
    const columnsMap = {}

    const pColumns = columns.map((column, index) => {
      let pColumn = null
      if (_.isString(column)) {
        pColumn = { id: column, accessor: column }
      } else {
        pColumn = { ...column, id: _.isUndefined(column.id) ? index.toString() : column.id }
      }
      columnsMap[pColumn.id] = pColumn
      return pColumn
    })

    return { pColumns, columnsMap }
  }, [columns])

  const [table, dispatch] = useReducer(reducer, { columns, isSorting, selectionType, initialSortState, initialFilterState, pageSize }, init)

  const toggleRowSelection = useCallback(id => value => {
    dispatch({
      type: 'TOGGLE-ROW-SELECTION',
      payload: {
        id,
        value
      }
    })
  }, [dispatch])

  const processedRows = useMemo(() => {
    return _.map(data, (row, index) => {
      const id = _.isUndefined(row.id) ? index.toString() : row.id
      const isSelected = table.selectionState?.rows[id] || false
      const selectionCell = table.selectionState
        ? {
            selectionCell: <SelectionCell selectionType={selectionType} isSelected={isSelected} disabled={row.disableSelection}
                                        onClick={() => toggleRowSelection(id)(!isSelected)}/>,
            cellProps: { onClick: selectionType === SELECTION_TYPE.SINGULAR && !row.disableSelection ? () => toggleRowSelection(id)(!isSelected) : null, ...row.cellProps }
          }
        : { ...row.cellProps }
      return {
        ...row,
        id,
        ...selectionCell
      }
    })
  }, [data, table, selectionType])

  const shownColumns = useMemo(() => {
    return pColumns.filter(column => {
      return !table.columnState[column.id]?.isHidden
    })
  }, [pColumns, table.columnState])

  const toggleSort = useCallback(({ id }) => {
    dispatch({
      type: 'TOGGLE-SORT',
      payload: {
        id
      }
    })
  }, [dispatch])

  const updateFilterState = useCallback((payload) => {
    dispatch({
      type: 'SET-FILTER',
      payload
    })
  }, [dispatch])

  const resetFilters = useCallback(() => {
    dispatch({
      type: 'RESET-FILTERS'
    })
  }, [dispatch])

  const updateRowState = useCallback(id => state => {
    dispatch({
      type: 'UPDATE-ROW',
      payload: {
        id,
        state
      }
    })
  }, [dispatch])

  const updateTextFilter = useCallback((value) => {
    dispatch({
      type: 'UPDATE-TEXT-FILTER',
      payload: {
        value
      }
    })
  }, [dispatch])

  const filteredRows = useMemo(() => {
    if ((table.filterState.count > 0 || table.filterState.textFilter) && !externalFilter) {
      return filter(table.filterState, processedRows, columns, columnsMap)
    } else {
      return processedRows
    }
  }, [table.filterState, processedRows, externalFilter, columns, columnsMap])

  const organizedRows = useMemo(() => {
    if (isSorting && !externalSort) {
      return sort(table.sortState, columnsMap, filteredRows)
    } else {
      return filteredRows
    }
  }, [isSorting, externalSort, filteredRows, table.sortState, columnsMap])

  const selectAllRows = useCallback(value => {
    dispatch({
      type: 'TOGGLE-SELECT-ALL',
      payload: {
        value,
        rows: organizedRows
      }
    })
  }, [dispatch, organizedRows])

  const pageChunks = useMemo(() => {
    if (pageSize) {
      return _.chunk(organizedRows, pageSize)
    } else {
      return null
    }
  }, [organizedRows, pageSize])

  const pagination = useMemo(() => {
    if (pageChunks) {
      const pageIndex = table.pageState.pageIndex
      const pageCount = pageChunks.length
      const gotoPage = (i) => {
        if (i >= 0 && (i < pageCount)) {
          dispatch({
            type: 'GO-TO-PAGE',
            payload: {
              index: i
            }
          })
        }
      }
      return {
        pageIndex,
        pageRows: pageChunks[pageIndex],
        pageCount,
        gotoPage
      }
    } else {
      return null
    }
  }, [pageChunks, table.pageState, dispatch])

  const processedColumns = useMemo(() => shownColumns.map(column => ({
    ...column,
    toggleSort: () => toggleSort({ id: column.id }),
    sortState: column.id === table.sortState?.column ? table.sortState?.sortOrder : null,
    filterState: table.filterState.columns[column.id],
    updateColumnFilter: (values) => updateFilterState({ id: column.id, values })
  })), [toggleSort, table.tempColumns, table.sortState, updateFilterState, table.filterState, shownColumns])

  const columnsWithSelection = useMemo(() => table.selectionState
    ? [{
        Header: '',
        accessor: 'selectionCell',
        id: 'selectionColumn',
        width: '60px'
      }, ...processedColumns]
    : processedColumns, [processedColumns])

  useRunOnceWhenTruthy(() => {
    defaultSelectedRows?.forEach(rowId => {
      toggleRowSelection(rowId)(true)
    })
  }, defaultSelectedRows && table.selectionState)

  return useMemo(() => ({
    rows: processedRows,
    organizedRows,
    columns: columnsWithSelection,
    sortState: table.sortState,
    filterState: table.filterState,
    tableState: table.tableState,
    rowState: table.rowState,
    selectionState: table.selectionState,
    toggleSort,
    updateFilterState,
    updateRowState,
    updateTextFilter,
    toggleRowSelection,
    selectAllRows,
    resetFilters,
    pagination,
    pageSize
  }), [table, organizedRows, toggleSort, updateFilterState, updateRowState, updateTextFilter, toggleRowSelection, selectAllRows, processedRows, resetFilters, pagination, columnsWithSelection])
}
